local MAJOR = "StarLibBuffer-1.0"
local MINOR = 13
assert(LibStub, MAJOR.." requires LibStub") 
local LibBuffer = LibStub:NewLibrary(MAJOR, MINOR)
if not LibBuffer then return end
local LibError = LibStub("StarLibError-1.0")
assert(LibError, MAJOR .. " requires StarLibError-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not LibBuffer.__index then
	LibBuffer.__index = LibBuffer
end

local new, del
do
	local pool = setmetatable({}, {__mode = "k"})
	function new(size)
		local obj
		if next(pool) then
			
			local high = 0
			for tbl in pairs(pool) do
				if not obj then obj = tbl end
				if #tbl >= size then
					obj = tbl
				end
			end
			pool[obj] = nil
		else
			obj = {}
		end
		
		return obj
	end
	function del(obj) 
		pool[obj] = true
	end
end
-- @name StarLibBuffer:New
-- @usage StarLibBuffer:New(name, size, defval, errorLevel)
-- @param name A name for the buffer
-- @param size The buffer's size
-- @param defval The buffer's default value. The alpha channel.
-- @param errorLevel The verbocity level
-- @return A new StarLibBuffer object

function LibBuffer:New(name, size, defval, errorLevel)
	assert(type(name) ~= "nil", "LibBuffer requires a name")
	assert(type(size) ~= "nil", "Buffer size is required")
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	obj.buffer = new(size)
	
	setmetatable(obj, self)

	obj.name = name
	obj.defval = defval or " "
	obj.error = LibError:New(MAJOR .. " " .. name, errorLevel)
	obj:Resize(size)

	return obj	
	
end

-- @name StarLibBuffer:Del
-- @usage StarLibBuffer:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LibBuffer:Del(buff)
	if not buff then buff = self end
	pool[buff] = true
	buff.error:Del()
	for i = 0, buff:Size() do
		buff.buffer[i] = false
	end
	buff.size = nil
	buff:Wipe()
	del(buff.buffer)
end

-- @name :Wipe
-- @usage :Wipe()
-- @return Nothing
function LibBuffer:Wipe()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

-- @name :Fill
-- @usage :Fill(val)
-- @param val The value to fill the buffer with
-- @return Nothing
function LibBuffer:Fill(val)
	for i = 0, self:Size() - 1 do
		self.buffer[i] = val
	end
end

-- @name :Resize
-- @usage :Resize(size)
-- @param size The buffer's new size
-- @return Nothing
function LibBuffer:Resize(size)
	assert(self.defval ~= nil)
	assert(type(size) == "number")
	if self:Size() < size then
		for i = self:Size(), size - 1 do
			self.buffer[i] = self.defval
		end
	end
	
	if self:Size() > size then
		for i = size - 1, self:Size() do
			self.buffer[i] = false
		end
	end
	
	self.size = size
	
	if size > 0 then
		for i = 0, size - 1 do
			assert(self.buffer[i] ~= nil, format("resize %d %d", i, size))
		end
	end
end

function LibBuffer:Replace(pos, val)
	assert(type(pos) == "number")
	assert(val ~= nil)
	if pos > self.size then self.size = pos end
	if type(val) ~= "string" then
		self.buffer[pos] = val
	else
		self.buffer[pos] = val:sub(1, 1)
	end
end

function LibBuffer:Clear()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

function LibBuffer:AsString()
	assert(type(self.defval) == "string")
	local str = ""
	for i = 0, self:Size() - 1 do
		if self.buffer[i] == false then break end
		if type(self.buffer[i]) == "string" then
			str = str .. self.buffer[i]
		else
			str = str .. self.defval
			self.error:Print(":AsString() - Bad index: " .. i .. ", type: " .. type(self.buffer[i]) .. ", size: " .. self:Size())
		end
	end
	return str
end

function LibBuffer:FromString(str)
	if type(str) ~= "string" then return end
	for i = 0, strlen(str) - 1 do
		self.buffer[i] = str:sub(i + 1, i + 1)
	end
	if strlen(str) < self:Size() then
		for i = strlen(str) - 1, self:Size() do
			self.buffer[i] = false
		end
	end
	self.size = strlen(str)
end

function LibBuffer:Size()
	--[[
	local count = 0
	for k in pairs(self.buffer) do
		count = count + 1
	end
	return count
	]]
	return self.size or 0
end

function LibBuffer:Sub(num1, num2)

	if not num2 then num2 = self:Size() - 1 end
	local t = LibBuffer:New(self.name .. ":" , num2 - num1 + 1, self.defval, self.errorLevel)
	for i = num1, num2 do
		assert(self.buffer[i], format("%s %d %d", self.name, i, self:Size()))
		t.buffer[i - num1] = self.buffer[i]
	end
	return t
end

function LibBuffer:Copy(other)
	table.wipe(self.buffer)
	for k, v in pairs(other.buffer) do
		self.buffer[k] = other.buffer[k]
	end
end

function LibBuffer:Merge(from, pos, len)
	local i = 0
	while i < len - (self.size - len) do
		assert(from.buffer[i] ~= nil)
		self.buffer[i + pos] = from.buffer[i]
		i = i + 1
	end
end

function LibBuffer:Memset(pos, char, length)
	for n = pos, pos + length - 1 do
		self.buffer[n] = char
	end
end

function LibBuffer:Memcopy(pos1, buffer, pos2, length)
	for n = 0, length - 1 do
		self.buffer[pos1 + n] = buffer.buffer[pos2 + n] or self.defval
	end
end
